import tester.*;

// PROGRAM BY DESIGN: HOW TO DESIGN CLASSES (REACH_JAVA)
// copyright 2011 VIERA KRNANOVA PROULX
// ISSEP 2011 WORKSHOP


// Part 9: A self-referential union of classes: representing lists
//         Abstracting over functional behavior
// File:   Book9.java

/*
            +---------+                
            | ILoBook4 |<--------------+
            +---------+               |
            +---------+               |
                / \                   |
                ---                   |
                 |                    |
       -------------------            |
       |                 |            |
  +----------+    +--------------+    |
  | MTLoBook4 |    | ConsLoBook4   |    |
  +----------+    +--------------+    |
  +----------+  +-| Book9 first   |    |
                | | ILoBook4 rest |----+
                | +--------------+ 
                |
                v
       +---------------+
       | Book9         |
       +---------------+
       | String title  |
       | String author |
       | int year      |
       +---------------+
*/

// to represent a list of books
abstract class ILoBook4 {

  // produce a list sorted in the given order from this list of books
  abstract ILoBook4 sort(BookBefore order);

  // insert the given book into this sorted list of books 
  // - by the given order
  abstract ILoBook4 insert(Book9 b, BookBefore order);

 // produce a list sorted by the title from this list of books
  ILoBook4 sortByTitle(){
    return this.sort(new BookByTitle());
  }

  // insert the given book into this sorted list of books - by title
  ILoBook4 insertByTitle(Book9 b){
    return this.insert(b, new BookByTitle());
  }

 // produce a list sorted by the year from this list of books
  ILoBook4 sortByYear(){
    return this.sort(new BookByYear());
  }

  // insert the given book into this sorted list of books - by year
  ILoBook4 insertByYear(Book9 b){
    return this.insert(b, new BookByYear());
  }

  // is this list of books is the same as the given one
  abstract boolean same(ILoBook4 that);

  // is this list of books is the same 
  // as the given empty list of books
  abstract boolean sameMTLoBook4(MTLoBook4 that);

  // is this list of books is the same 
  // as the given nonempty list of books
  abstract boolean sameConsLoBook4(ConsLoBook4 that);
}

// to represent an empty list of books
class MTLoBook4 extends ILoBook4 {

  MTLoBook4() {
  }

  // produce a list sorted by the given order from this list of books
  ILoBook4 sort(BookBefore order){ return this; }

  // insert the given book into this sorted list of books 
  // - by the given order
  ILoBook4 insert(Book9 b, BookBefore order){ 
    return new ConsLoBook4(b, this);
  }

  // is this list of books is the same as the given one
  boolean same(ILoBook4 that){
    return that.sameMTLoBook4(this);
  }

  // is this list of books is the same 
  // as the given empty list of books
  boolean sameMTLoBook4(MTLoBook4 that){ return true; }

  // is this list of books is the same
  // as the given nonempty list of books
  boolean sameConsLoBook4(ConsLoBook4 that){ return false; }
}

// to represent a nonempty list of Books
class ConsLoBook4 extends ILoBook4 {
  Book9 first;
  ILoBook4 rest;

  ConsLoBook4(Book9 first, ILoBook4 rest) {
    this.first = first;
    this.rest = rest;
  }

  /* TEMPLATE:
    ... this.first ...                       -- Book9
    ... this.rest ...                        -- ILoBook4
    ... this.rest.sort(order) ...            -- ILoBook4 (sorted)
    ... this.rest.insert(Book, order) ...    -- ILoBook4 (sorted)

    ... this.rest.sortByTitle() ...          -- ILoBook4 (sorted)
    ... this.rest.insertByTitle(Book) ...    -- ILoBook4 (sorted)
    ... this.rest.sortByYear() ...           -- ILoBook4 (sorted)
    ... this.rest.insertByYear(Book) ...     -- ILoBook4 (sorted)
  */

  // produce a list sorted by the given order from this list of books
  ILoBook4 sort(BookBefore order){  
    return this.rest.sort(order).insert(this.first, order);
  }

  // insert the given book into this sorted list of books 
  // - by the given order
  ILoBook4 insert(Book9 b, BookBefore order){
    if (order.isBefore(this.first, b))
      return new ConsLoBook4(this.first, this.rest.insert(b, order));
    else
      return new ConsLoBook4(b, this);
  }

  // to determine whether this list of books is the same as the given one
  boolean same(ILoBook4 that){
    return that.sameConsLoBook4(this);
  }

  // is this list of books is the same as the given empty list of books
  boolean sameMTLoBook4(MTLoBook4 that){ return false; }

  // is this list of books is the same as the given nonempty list of books
  boolean sameConsLoBook4(ConsLoBook4 that){ 
    return this.first.same(that.first) &&
           this.rest.same(that.rest);
  }
}

// to represent a book
class Book9 {
  String title;
  String author;
  int year;

  Book9(String title, String author, int year) {
    this.title = title;
    this.author = author;
    this.year = year;
  }
/*
  // ** DRAFT TEMPLATE ** Edit as needed.
  ??? mmm() {
    ... this.title ...    -- String
    ... this.author ...   -- String
    ... this.year ...     -- int
  }
*/

  // is the title of this book before the title of the given book
  boolean beforeByTitle(Book9 that){
    return this.title.compareTo(that.title) < 0;
  }

  // is the year of this book before the year of the given book
  boolean beforeByYear(Book9 that){
    return this.year < that.year;
  }

  // to determine whether this book is the same as the given book
  boolean same(Book9 that){
    return this.title.equals(that.title) &&
           this.author.equals(that.author) &&
           this.year == that.year;
  }
}

// to represent a comparison operation for books
interface BookBefore{
  // is the first book before the second one in this ordering?
  boolean isBefore(Book9 book1, Book9 book2);
}

// function to determine the book order by the title
class BookByTitle implements BookBefore{
  // is the first book before the second one ordered by the title
  public boolean isBefore(Book9 book1, Book9 book2){
    return book1.title.compareTo(book2.title) < 0;
  }  
}

// function to determine the book order by the year
class BookByYear implements BookBefore{
  // is the first book before the second one ordered by the year
  public boolean isBefore(Book9 book1, Book9 book2){
    return book1.year < book2.year;
  }  
}

// examples of data and tests
class ExamplesBook9{
  ExamplesBook9() {}

  Book9 b1 = new Book9("RC", "DD", 1796);
  Book9 b2 = new Book9("LPP", "StEx", 1942);
  Book9 b3 = new Book9("HtDP", "FFFK", 2001);

  ILoBook4 mtbooks = new MTLoBook4();
  ILoBook4 books1 = new ConsLoBook4(this.b1, 
                   new ConsLoBook4(this.b2, 
                   this.mtbooks));
  ILoBook4 books2 = new ConsLoBook4(this.b3, this.books1);

  ILoBook4 bSorted = new ConsLoBook4(this.b3, 
                    new ConsLoBook4(this.b1, 
                    this.mtbooks));

  ILoBook4 allSorted = new ConsLoBook4(this.b3, 
                      new ConsLoBook4(this.b2, 
                      new ConsLoBook4(this.b1, 
                      this.mtbooks)));


  // tests for the method beforeByTitle in the class Book
  boolean testBeforeByTitle(Tester t){
	  return
	  t.checkExpect(this.b2.beforeByTitle(this.b1), true) &&
	  t.checkExpect(this.b2.beforeByTitle(this.b3), false);
  }

  // tests for the method insertByTitle for ILoBook4 classes
  boolean testInsertByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.insertByTitle(this.b1),
          new ConsLoBook4(this.b1, this.mtbooks)) &&
	  t.checkExpect(this.bSorted.insertByTitle(this.b2),
          new ConsLoBook4(this.b3,
            new ConsLoBook4(this.b2,
              new ConsLoBook4(this.b1,
                this.mtbooks))));
  }

  // tests for the method sortByTitle for ILoBook4 classes
  boolean testSortByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.sortByTitle(), this.mtbooks) &&
	  t.checkExpect(this.books2.sortByTitle(), this.allSorted);
  }

  // tests for the method beforeByYear in the class Book
  boolean testBeforeByYear(Tester t){
	  return
	  t.checkExpect(this.b2.beforeByYear(this.b1), false) &&
	  t.checkExpect(this.b2.beforeByYear(this.b3), true);
  }

  // tests for the method insertByYear for ILoBook4 classes
  boolean testInsertByYear(Tester t){
	  return
	  t.checkExpect(this.mtbooks.insertByYear(this.b1),
          new ConsLoBook4(this.b1, this.mtbooks)) &&
	  t.checkExpect(this.books1.insertByYear(this.b3),
          new ConsLoBook4(this.b1,
            new ConsLoBook4(this.b2,
              new ConsLoBook4(this.b3,
                this.mtbooks)))); 
  }

// tests for the method sortByYear for ILoBook4 classes
  boolean testSortByYear(Tester t){
	  return
	  t.checkExpect(this.mtbooks.sortByYear(), this.mtbooks) &&
	  t.checkExpect(this.books2.sortByYear(),
          new ConsLoBook4(this.b1, 
            new ConsLoBook4(this.b2, 
              new ConsLoBook4(this.b3, 
                this.mtbooks))));
  }

  // run the tests
  public static void main(String[] argv){
	  ExamplesBook9 eb9 = new ExamplesBook9();
	  Tester.runFullReport(eb9);
  }  
}
